"""Implementation of NoFreeFormRule."""

from __future__ import annotations

import functools
import re
import sys
from typing import TYPE_CHECKING, Any

from ansible.errors import AnsibleParserError
from ansible.parsing.splitter import split_args
from ruamel.yaml.scalarstring import DoubleQuotedScalarString, SingleQuotedScalarString

from ansiblelint.constants import INCLUSION_ACTION_NAMES
from ansiblelint.rules import AnsibleLintRule, TransformMixin
from ansiblelint.rules.key_order import task_property_sorter

if TYPE_CHECKING:
    from ruamel.yaml.comments import CommentedMap, CommentedSeq

    from ansiblelint.errors import MatchError
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class NoFreeFormRule(AnsibleLintRule, TransformMixin):
    """Rule for detecting discouraged free-form syntax for action modules."""

    id = "no-free-form"
    description = "Avoid free-form inside files as it can produce subtle bugs."
    severity = "MEDIUM"
    tags = ["syntax", "risk"]
    version_changed = "6.8.0"
    needs_raw_task = True
    cmd_shell_re = re.compile(
        r"(chdir|creates|executable|removes|stdin|stdin_add_newline|warn)=",
    )
    _ids = {
        "no-free-form[raw]": "Avoid embedding `executable=` inside raw calls, use explicit args dictionary instead.",
        "no-free-form[raw-non-string]": "Passing a non string value to `raw` module is neither documented or supported.",
    }

    @staticmethod
    def _has_unmatched_quote(val: str) -> bool:
        """Detect values starting with a quote but lacking a closing match."""
        if not val:
            return False
        if val[0] in "\"'":
            if len(val) == 1:
                return True
            return val[-1] != val[0]
        return False

    @staticmethod
    def _normalize_value(val: str) -> Any:
        """Normalize quoted values while keeping Jinja/spacing intact."""
        if len(val) >= 2 and val[0] == val[-1] and val[0] in "\"'":
            return (
                DoubleQuotedScalarString if val[0] == '"' else SingleQuotedScalarString
            )(val[1:-1])
        return val

    @classmethod
    def _parse_module_opts(cls, value: str) -> dict[str, Any]:
        """Parse module opts, falling back to cmd on unbalanced quotes."""
        try:
            parts = split_args(value)  # type: ignore[no-untyped-call]
        except AnsibleParserError:
            return {"cmd": value}
        module_opts: dict[str, Any] = {}
        cmd_parts: list[str] = []
        for part in parts:
            if "=" in part:
                key, raw_value = part.split("=", 1)
                if cls._has_unmatched_quote(raw_value):
                    return {"cmd": value}
                module_opts[key] = cls._normalize_value(raw_value)
            else:
                cmd_parts.append(part)
        if cmd_parts:
            module_opts["cmd"] = " ".join(cmd_parts)
        return module_opts

    @classmethod
    def _parse_raw_value(cls, value: str) -> tuple[str, dict[str, Any]]:
        """Parse raw module string, extracting executable when possible."""
        try:
            parts = split_args(value)  # type: ignore[no-untyped-call]
        except AnsibleParserError:
            return value, {}
        exec_key_val: dict[str, Any] = {}
        raw_cmd_parts: list[str] = []
        for part in parts:
            if part.startswith("executable="):
                _, raw_value = part.split("=", 1)
                exec_key_val["executable"] = cls._normalize_value(raw_value)
            else:
                raw_cmd_parts.append(part)
        return " ".join(raw_cmd_parts), exec_key_val

    @staticmethod
    def _sorted_module_opts(module_opts: dict[str, Any]) -> dict[str, Any]:
        sorted_module_opts: dict[str, Any] = {}
        for key in sorted(
            module_opts.keys(),
            key=functools.cmp_to_key(task_property_sorter),
        ):
            sorted_module_opts[key] = module_opts[key]
        return sorted_module_opts

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> list[MatchError]:
        results: list[MatchError] = []
        action = task["action"]["__ansible_module_original__"]

        if action in INCLUSION_ACTION_NAMES:
            return results

        action_value = task["__raw_task__"].get(action, None)
        if task["action"].get("__ansible_module__", None) == "raw":
            if isinstance(action_value, str):
                if "executable=" in action_value:
                    results.append(
                        self.create_matcherror(
                            message="Avoid embedding `executable=` inside raw calls, use explicit args dictionary instead.",
                            lineno=task.line,
                            filename=file,
                            tag=f"{self.id}[raw]",
                        ),
                    )
            else:
                results.append(
                    self.create_matcherror(
                        message="Passing a non string value to `raw` module is neither documented or supported.",
                        lineno=task.line,
                        filename=file,
                        tag=f"{self.id}[raw-non-string]",
                    ),
                )
        elif isinstance(action_value, str) and "=" in action_value:
            fail = False
            if task["action"].get("__ansible_module__") in (
                "ansible.builtin.command",
                "ansible.builtin.shell",
                "ansible.windows.win_command",
                "ansible.windows.win_shell",
                "command",
                "shell",
                "win_command",
                "win_shell",
            ):
                if self.cmd_shell_re.search(action_value):
                    fail = True
            else:
                fail = True
            if fail:
                results.append(
                    self.create_matcherror(
                        message=f"Avoid using free-form when calling module actions. ({action})",
                        lineno=task.line,
                        filename=file,
                        details=action,
                    ),
                )
        return results

    def transform(
        self,
        match: MatchError,
        lintable: Lintable,
        data: CommentedMap | CommentedSeq | str,
    ) -> None:
        if "no-free-form" in match.tag:
            task = self.seek(match.yaml_path, data)

            if match.tag == "no-free-form":
                target_module = match.details

                for _ in range(len(task)):
                    k, v = task.popitem(False)
                    # identify module as key and process its value
                    if k == target_module and isinstance(v, str):
                        module_opts = self._parse_module_opts(v)
                        task[k] = self._sorted_module_opts(module_opts)
                    else:
                        task[k] = v

                match.fixed = True
            elif match.tag == "no-free-form[raw]":
                for _ in range(len(task)):
                    k, v = task.popitem(False)
                    if isinstance(v, str) and "executable" in v:
                        raw_value, exec_key_val = self._parse_raw_value(v)
                        task[k] = raw_value
                        if exec_key_val:
                            task["args"] = exec_key_val
                    else:
                        task[k] = v
                match.fixed = True


if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("file", "expected"),
        (
            pytest.param("examples/playbooks/rule-no-free-form-pass.yml", 0, id="pass"),
            pytest.param("examples/playbooks/rule-no-free-form-fail.yml", 6, id="fail"),
        ),
    )
    def test_rule_no_free_form(
        default_rules_collection: RulesCollection,
        file: str,
        expected: int,
    ) -> None:
        """Validate that rule works as intended."""
        results = Runner(file, rules=default_rules_collection).run()

        rule_results = [r for r in results if r.rule.id == NoFreeFormRule.id]

        for result in rule_results:
            assert result.rule.id == NoFreeFormRule.id, result
        assert len(rule_results) == expected

    def test_no_free_form_transform_error_handling() -> None:
        """Test that transform handles malformed quoted strings."""
        from ruamel.yaml.comments import CommentedMap

        from ansiblelint.errors import MatchError

        rule = NoFreeFormRule()
        task = CommentedMap({"ansible.builtin.shell": 'chdir=" /tmp echo foo'})
        match = MatchError(
            message="test",
            rule=rule,
            details="ansible.builtin.shell",
            tag="no-free-form",
        )

        rule.transform(match, None, task)  # type: ignore[arg-type]
        assert task["ansible.builtin.shell"] == {"cmd": 'chdir=" /tmp echo foo'}

    def test_no_free_form_transform_jinja_with_spaces() -> None:
        """Test that Jinja expressions with spaces are preserved."""
        from ruamel.yaml.comments import CommentedMap

        from ansiblelint.errors import MatchError

        rule = NoFreeFormRule()
        task = CommentedMap(
            {"ansible.builtin.dnf": "name={{ item }} state=latest"},
        )
        match = MatchError(
            message="test",
            rule=rule,
            details="ansible.builtin.dnf",
            tag="no-free-form",
        )

        rule.transform(match, None, task)  # type: ignore[arg-type]
        assert task["ansible.builtin.dnf"]["name"] == "{{ item }}"
        assert task["ansible.builtin.dnf"]["state"] == "latest"
        assert "cmd" not in task["ansible.builtin.dnf"]

    def test_no_free_form_transform_unmatched_quote_value() -> None:
        """Test that malformed quoted values fall back to cmd."""
        from ruamel.yaml.comments import CommentedMap

        from ansiblelint.errors import MatchError

        rule = NoFreeFormRule()
        task = CommentedMap(
            {"ansible.builtin.command": 'name="foo"bar chdir=/tmp'},
        )
        match = MatchError(
            message="test",
            rule=rule,
            details="ansible.builtin.command",
            tag="no-free-form",
        )

        rule.transform(match, None, task)  # type: ignore[arg-type]
        assert task["ansible.builtin.command"] == {
            "cmd": 'name="foo"bar chdir=/tmp',
        }

    def test_no_free_form_transform_raw_unbalanced_executable() -> None:
        """Test raw transform fallback when executable value is unbalanced."""
        from ruamel.yaml.comments import CommentedMap

        from ansiblelint.errors import MatchError

        rule = NoFreeFormRule()
        task = CommentedMap(
            {"ansible.builtin.raw": 'executable="/bin/bash echo foo'},
        )
        match = MatchError(
            message="test",
            rule=rule,
            tag="no-free-form[raw]",
        )

        rule.transform(match, None, task)  # type: ignore[arg-type]
        assert task["ansible.builtin.raw"] == 'executable="/bin/bash echo foo'
        assert "args" not in task



"""PyTest Fixtures.

They should not be imported, instead add code below to your root conftest.py
file:

pytest_plugins = ['ansiblelint.testing']
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from ansiblelint.app import App, get_app
from ansiblelint.config import Options
from ansiblelint.constants import DEFAULT_RULESDIR
from ansiblelint.rules import RulesCollection
from ansiblelint.testing import RunFromText

if TYPE_CHECKING:
    from _pytest.fixtures import SubRequest

# pylint: disable=redefined-outer-name


# The sessions scope does not apply to xdist, so we will still have one
# session for each worker, but at least it will a limited number.
@pytest.fixture(name="default_rules_collection", scope="session")
def fixture_default_rules_collection(app: App) -> RulesCollection:
    """Return default rule collection."""
    assert DEFAULT_RULESDIR.is_dir()
    config_options = Options()
    config_options.enable_list = ["no-same-owner"]
    # That is instantiated very often and do want to avoid ansible-galaxy
    # install errors due to concurrency.
    config_options.offline = True
    return RulesCollection(
        app=app, rulesdirs=[DEFAULT_RULESDIR], options=config_options
    )


@pytest.fixture
def empty_rule_collection(app: App) -> RulesCollection:
    """Return an empty rules collection."""
    return RulesCollection(app=app)


@pytest.fixture
def default_text_runner(
    default_rules_collection: RulesCollection, app: App
) -> RunFromText:
    """Return RunFromText instance for the default set of collections."""
    return RunFromText(default_rules_collection, app)


@pytest.fixture
def rule_runner(request: SubRequest, app: App) -> RunFromText:
    """Return runner for a specific rule class."""
    rule_class = request.param
    config_options = Options()
    config_options.enable_list.append(rule_class().id)
    collection = RulesCollection(app=app, options=config_options)
    collection.register(rule_class())
    return RunFromText(collection, app)


@pytest.fixture(name="config_options")
def fixture_config_options() -> Options:
    """Return configuration options that will be restored after testrun."""
    return Options()


@pytest.fixture(scope="session")
def app() -> App:
    """Return the application instance."""
    return get_app(offline=True)



#!/usr/bin/env python
# Copyright (c) 2013-2014 Will Thames <will@thames.id.au>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Command line implementation."""

from __future__ import annotations

import errno
import logging
import os
import pathlib
import shutil
import site
import sys
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

from ansible_compat.prerun import get_cache_dir
from filelock import BaseFileLock, FileLock, Timeout

from ansiblelint.constants import RC, SKIP_SCHEMA_UPDATE

# safety check for broken ansible core, needs to happen first
try:
    # pylint: disable=unused-import
    from ansible.parsing.dataloader import DataLoader  # noqa: F401

except Exception as _exc:  # pylint: disable=broad-exception-caught # noqa: BLE001
    logging.fatal(_exc)
    sys.exit(RC.INVALID_CONFIG)
# pylint: disable=ungrouped-imports
from ansiblelint import cli
from ansiblelint._mockings import _perform_mockings_cleanup
from ansiblelint.app import get_app
from ansiblelint.config import (
    Options,
    get_deps_versions,
    get_version_warning,
    log_entries,
    options,
)
from ansiblelint.loaders import IgnoreRule, IgnoreRuleQualifier, load_ignore_txt
from ansiblelint.output import (
    console,
    console_stderr,
    reconfigure,
    render_yaml,
    should_do_markup,
)
from ansiblelint.runner import get_matches
from ansiblelint.skip_utils import normalize_tag
from ansiblelint.version import __version__

if TYPE_CHECKING:
    # RulesCollection must be imported lazily or ansible gets imported too early.

    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import LintResult


_logger = logging.getLogger(__name__)


class LintLogHandler(logging.Handler):
    """Custom handler that uses our rich stderr console."""

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            console_stderr.print(f"[dim]{msg}[/]")
        except RecursionError:  # See issue 36272
            raise
        except Exception:  # pylint: disable=broad-exception-caught # noqa: BLE001
            self.handleError(record)


def initialize_logger(level: int = 0) -> None:
    """Set up the global logging level based on the verbosity number."""
    # We are about to act on the root logger, which defaults to logging.WARNING.
    # That is where our 0 (default) value comes from.
    verbosity_map = {
        -2: logging.CRITICAL,
        -1: logging.ERROR,
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }

    handler = LintLogHandler()
    formatter = logging.Formatter("%(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    logger = logging.getLogger()
    logger.addHandler(handler)
    # Unknown logging level is treated as DEBUG
    logging_level = verbosity_map.get(level, logging.DEBUG)
    logger.setLevel(logging_level)
    logging.captureWarnings(True)  # pass all warnings.warn() messages through logging
    # Use module-level _logger instance to validate it
    _logger.debug("Logging initialized to level %s", logging_level)


def initialize_options(arguments: list[str] | None = None) -> BaseFileLock | None:
    """Load config options and store them inside options module."""
    cache_dir_lock = None
    new_options = cli.get_config(arguments or [])
    new_options.cwd = pathlib.Path.cwd()

    if new_options.colored is None:
        new_options.colored = should_do_markup()

    # persist loaded configuration inside options module
    for k, v in new_options.__dict__.items():
        setattr(options, k, v)

    # rename deprecated ids/tags to newer names
    options.tags = [normalize_tag(tag) for tag in options.tags]
    options.skip_list = [normalize_tag(tag) for tag in options.skip_list]
    options.warn_list = [normalize_tag(tag) for tag in options.warn_list]

    options.configured = True
    if not (
        options.version
        or options.list_profiles
        or options.list_rules
        or options.list_tags
    ):
        options.cache_dir = get_cache_dir(pathlib.Path(options.project_dir))

    options.project_dir = Path(options.project_dir).resolve().as_posix()

    # add a lock file so we do not have two instances running inside at the same time
    if options.cache_dir:
        options.cache_dir.mkdir(parents=True, exist_ok=True)

        # lock file can only be used if cache_dir is set and writable
        if not options.offline:  # pragma: no cover
            cache_dir_lock = FileLock(
                f"{options.cache_dir}/.lock",
            )
            try:
                cache_dir_lock.acquire(timeout=180)
            except Timeout:  # pragma: no cover
                _logger.error(  # noqa: TRY400
                    "Timeout waiting for another instance of ansible-lint to release the lock.",
                )
                sys.exit(RC.LOCK_TIMEOUT)

    # Avoid extra output noise from Ansible about using devel versions
    if "ANSIBLE_DEVEL_WARNING" not in os.environ:  # pragma: no branch
        os.environ["ANSIBLE_DEVEL_WARNING"] = "false"

    return cache_dir_lock


def _do_list(rules: RulesCollection) -> int:
    # On purpose lazy-imports to avoid pre-loading Ansible
    # pylint: disable=import-outside-toplevel
    from ansiblelint.generate_docs import rules_as_str

    if options.list_rules:
        console.print(
            rules_as_str(rules),
        )
        return 0

    if options.list_tags:
        console.print(render_yaml(rules.list_tags()))
        return 0

    # we should not get here!
    return 1


# noinspection PyShadowingNames
def _do_transform(result: LintResult, opts: Options) -> None:
    """Create and run Transformer."""
    if "yaml" in opts.skip_list:
        # The transformer rewrites yaml files, but the user requested to skip
        # the yaml rule or anything tagged with "yaml", so there is nothing to do.
        return

    # On purpose lazy-imports to avoid loading transforms unless requested
    # pylint: disable=import-outside-toplevel
    from ansiblelint.transformer import Transformer

    transformer = Transformer(result, options)

    # this will mark any matches as fixed if the transforms repaired the issue
    transformer.run()


def support_banner() -> None:
    """Display support banner when running on unsupported platform."""


def fix(runtime_options: Options, result: LintResult, rules: RulesCollection) -> None:
    """Fix the linting errors.

    :param options: Options object
    :param result: LintResult object
    """
    match_count = len(result.matches)
    _logger.debug("Begin fixing: %s matches", match_count)
    ruamel_safe_version = "0.17.26"

    # pylint: disable=import-outside-toplevel
    from packaging.version import Version
    from ruamel.yaml import __version__ as ruamel_yaml_version_str

    # pylint: enable=import-outside-toplevel

    if Version(ruamel_safe_version) > Version(
        ruamel_yaml_version_str
    ):  # pragma: no cover
        _logger.warning(
            "We detected use of `--fix` feature with a buggy ruamel-yaml %s library instead of >=%s, upgrade it before reporting any bugs like dropped comments.",
            ruamel_yaml_version_str,
            ruamel_safe_version,
        )
    acceptable_tags = {"all", "none", *rules.known_transform_tags()}
    unknown_tags = set(options.write_list).difference(acceptable_tags)

    if unknown_tags:  # pragma: no cover
        _logger.error(
            "Found invalid value(s) (%s) for --fix arguments, must be one of: %s. Valid values are limited by the configured profile.",
            ", ".join(unknown_tags),
            ", ".join(acceptable_tags),
        )
        sys.exit(RC.INVALID_CONFIG)
    _do_transform(result, options)

    rerun = ["yaml"]
    resolved = []
    for idx, match in reversed(list(enumerate(result.matches))):
        _logger.debug("Fixing: (%s of %s) %s", match_count - idx, match_count, match)
        if match.fixed:
            _logger.debug("Fixed, removed: %s", match)
            result.matches.pop(idx)
            continue
        if match.rule.id not in rerun:
            _logger.debug("Not rerun eligible: %s", match)
            continue

        uid = (match.rule.id, match.filename)
        if uid in resolved:
            _logger.debug("Previously resolved: %s", match)
            result.matches.pop(idx)
            continue
        _logger.debug("Rerunning: %s", match)
        runtime_options.tags = [match.rule.id]
        runtime_options.lintables = [match.filename]
        runtime_options._skip_ansible_syntax_check = True  # noqa: SLF001
        new_results = get_matches(rules, runtime_options)
        if not new_results.matches:
            _logger.debug("Newly resolved: %s", match)
            result.matches.pop(idx)
            resolved.append(uid)
            continue
        if match in new_results.matches:
            _logger.debug("Still found: %s", match)
            continue
        _logger.debug("Fixed, removed: %s", match)
        result.matches.pop(idx)


# By default, matches ignored in .ansible-lint-ignore are treated
# as warnings [1].  If the user explicitly adds a skip qualifier
# to the rule, it is treated as skipped here and does not show up
# even as a warning.
# [1] https://github.com/ansible/ansible-lint/issues/3068
def _rule_is_skipped(tag: str, rules: set[IgnoreRule]) -> bool:
    for rule in rules:
        if tag != rule.rule:
            return False
        return IgnoreRuleQualifier.SKIP in rule.qualifiers
    return False


# pylint: disable=too-many-locals,too-many-statements
def main(argv: list[str] | None = None) -> int:
    """Linter CLI entry point."""
    must_exit = False
    # alter PATH if needed (venv support)
    path_inject(argv[0] if argv and argv[0] else "")

    if argv is None:  # pragma: no cover
        argv = sys.argv

    warnings.simplefilter(
        "ignore", ResourceWarning
    )  # suppress "enable tracemalloc to get the object allocation traceback"
    with warnings.catch_warnings(record=True) as warns:
        # do not use "ignore" as we will miss to collect them
        warnings.simplefilter(action="default")

        cache_dir_lock = initialize_options(argv[1:])

        reconfigure(colored=options.colored)

        if options.version:
            deps = get_deps_versions()
            msg = f"ansible-lint [repr.number]{__version__}[/] using[dim]"
            for k, v in deps.items():
                msg += f" {k}:[repr.number]{v}[/]"
            msg += "[/]"
            console.print(msg)
            msg = get_version_warning()
            if msg:  # pragma: no cover
                console.print(msg)
            support_banner()
            must_exit = True
        else:
            support_banner()

        initialize_logger(options.verbosity)
        for level, message in log_entries:
            _logger.log(level, message)
        _logger.debug("Options: %s", options)
        _logger.debug("CWD: %s", Path.cwd())

    for warn in warns:  # pragma: no cover
        _logger.warning(str(warn.message))
    warnings.resetwarnings()

    if must_exit:
        sys.exit(0)
    # checks if we have `ANSIBLE_LINT_SKIP_SCHEMA_UPDATE` set to bypass schema
    # update. Also skip if in offline mode.
    # env var set to skip schema refresh
    skip_schema_update = (
        bool(
            int(
                os.environ.get(
                    SKIP_SCHEMA_UPDATE,
                    "0",
                ),
            ),
        )
        or options.offline
        or options.nodeps
    )

    if not skip_schema_update:  # pragma: no cover
        # pylint: disable=import-outside-toplevel
        from ansiblelint.schemas.__main__ import refresh_schemas

        refresh_schemas()

    # pylint: disable=import-outside-toplevel
    from ansiblelint.rules import RulesCollection

    if options.list_profiles:
        from ansiblelint.generate_docs import profiles_as_md

        profiles_as_md().display()
        return 0

    app = get_app(
        offline=None,
        cached=True,
    )  # to be sure we use the offline value from settings
    rules = RulesCollection(
        app=app,
        rulesdirs=options.rulesdirs,
        profile_name=options.profile,
        options=options,
    )

    if options.list_rules or options.list_tags:
        return _do_list(rules)

    if isinstance(options.tags, str):
        options.tags = options.tags.split(",")  # pragma: no cover
    result = get_matches(rules, options)

    mark_as_success = True

    if options.strict and result.matches:
        mark_as_success = False

    # Remove skip_list items from the result
    result.matches = [m for m in result.matches if m.tag not in app.options.skip_list]
    # load ignore file
    ignore_map = load_ignore_txt(options.ignore_file)
    # prune qualified skips from ignore file
    result.matches = [
        m for m in result.matches if not _rule_is_skipped(m.tag, ignore_map[m.filename])
    ]
    # others entries are ignored
    for match in result.matches:
        if match.tag in [
            i.rule for i in ignore_map[match.filename]
        ]:  # pragma: no cover
            match.ignored = True
            _logger.debug("Ignored: %s", match)

    if app.yamllint_config.incompatible:  # pragma: no cover
        _logger.log(
            level=logging.ERROR if options.write_list else logging.WARNING,
            msg=app.yamllint_config.incompatible,
        )

    if options.write_list:
        if app.yamllint_config.incompatible:  # pragma: no cover
            sys.exit(RC.INVALID_CONFIG)
        fix(runtime_options=options, result=result, rules=rules)

    app.render_matches(result.matches)

    _perform_mockings_cleanup(app.options)
    if cache_dir_lock:
        cache_dir_lock.release()
        pathlib.Path(cache_dir_lock.lock_file).unlink(missing_ok=True)
    if options.mock_filters:
        _logger.warning(
            "The following filters were mocked during the run: %s",
            ",".join(options.mock_filters),
        )

    return app.report_outcome(result, mark_as_success=mark_as_success)


def _run_cli_entrypoint() -> None:
    """Invoke the main entrypoint with current CLI args.

    This function also processes the runtime exceptions.
    """
    try:
        sys.exit(main(sys.argv))
    except OSError as exc:
        # NOTE: Only "broken pipe" is acceptable to ignore
        if exc.errno != errno.EPIPE:  # pragma: no cover
            raise
    except KeyboardInterrupt:  # pragma: no cover
        sys.exit(RC.EXIT_CONTROL_C)
    except RuntimeError as exc:  # pragma: no cover
        raise SystemExit(exc) from exc


def path_inject(own_location: str = "") -> None:
    """Add python interpreter path to top of PATH to fix outside venv calling."""
    # This make it possible to call ansible-lint that was installed inside a
    # virtualenv without having to pre-activate it. Otherwise subprocess will
    # either fail to find ansible executables or call the wrong ones.
    #
    # This must be run before we do run any subprocesses, and loading config
    # does this as part of the ansible detection.
    paths = [x for x in os.environ.get("PATH", "").split(os.pathsep) if x]

    # Expand ~ in PATH as it known to break many tools
    expanded = False
    for idx, path in enumerate(paths):
        if path.startswith("~"):  # pragma: no cover
            paths[idx] = str(Path(path).expanduser())
            expanded = True
    if expanded:  # pragma: no cover
        print(  # noqa: T201
            "WARNING: PATH altered to expand ~ in it. Read https://stackoverflow.com/a/44704799/99834 and correct your system configuration.",
            file=sys.stderr,
        )

    inject_paths = []

    userbase_bin_path = Path(site.getuserbase()) / "bin"

    if (
        str(userbase_bin_path) not in paths
        and (userbase_bin_path / "bin" / "ansible").exists()
    ):
        inject_paths.append(userbase_bin_path.resolve().as_posix())

    py_path = Path(sys.executable).parent.resolve()
    pipx_path = os.environ.get("PIPX_HOME", "pipx")
    if (
        str(py_path) not in paths
        and (py_path / "ansible").exists()
        and pipx_path not in str(py_path)
    ):
        inject_paths.append(py_path.as_posix())

    # last option, if nothing else is found, just look next to ourselves...
    if own_location:
        own_location = os.path.realpath(own_location)
        parent = Path(own_location).parent
        if (parent / "ansible").exists() and str(parent) not in paths:
            inject_paths.append(str(parent))

    if not os.environ.get("PYENV_VIRTUAL_ENV", None):
        if inject_paths and not any((
            all("pipx" in p for p in inject_paths),
            all("uv/tools" in p for p in inject_paths),
        )):
            print(  # noqa: T201
                f"WARNING: PATH altered to include {', '.join(inject_paths)} :: This is usually a sign of broken local setup, which can cause unexpected behaviors.",
                file=sys.stderr,
            )
        if inject_paths or expanded:
            os.environ["PATH"] = os.pathsep.join([*inject_paths, *paths])

    # We do know that finding ansible in PATH does not guarantee that it is
    # functioning or that is in fact the same version that was installed as
    # our dependency, but addressing this would be done by ansible-compat.
    for cmd in ("ansible",):
        if not shutil.which(cmd):  # pragma: no cover
            msg = f"Failed to find runtime dependency '{cmd}' in PATH"
            raise RuntimeError(msg)


if __name__ == "__main__":
    _run_cli_entrypoint()



"""Optional Ansible-lint rule to warn use of run_once with strategy free."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from ansiblelint.rules import AnsibleLintRule

if TYPE_CHECKING:
    from ansiblelint.errors import MatchError
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class RunOnce(AnsibleLintRule):
    """Run once should use strategy other than free."""

    id = "run-once"
    link = "https://docs.ansible.com/projects/ansible/latest/reference_appendices/playbooks_keywords.html"
    description = "When using run_once, we should avoid using strategy as free."

    tags = ["idiom"]
    severity = "MEDIUM"
    version_changed = "6.12.0"
    _ids = {
        "run-once[task]": "Using run_once may behave differently if strategy is set to free.",
        "run-once[play]": "Play uses strategy: free",
    }

    def matchplay(self, file: Lintable, data: dict[str, Any]) -> list[MatchError]:
        """Return matches found for a specific playbook."""
        # If the Play uses the 'strategy' and it's value is set to free

        if not file or file.kind != "playbook" or not data:
            return []

        strategy: Any = data.get("strategy")
        run_once = data.get("run_once", False)
        if (not strategy and not run_once) or strategy != "free":
            return []
        lineno = getattr(strategy, "_line_number", 1)
        return [
            self.create_matcherror(
                message="Play uses strategy: free",
                filename=file,
                tag=f"{self.id}[play]",
                lineno=lineno,
            ),
        ]

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> list[MatchError]:
        """Return matches for a task."""
        if not file or file.kind != "playbook":
            return []

        run_once = task.get("run_once", False)
        if not run_once:
            return []
        return [
            self.create_matcherror(
                message="Using run_once may behave differently if strategy is set to free.",
                filename=file,
                tag=f"{self.id}[task]",
                lineno=task.line,
            ),
        ]


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("test_file", "failure"),
        (
            pytest.param("examples/playbooks/run-once-pass.yml", 0, id="pass"),
            pytest.param("examples/playbooks/run-once-fail.yml", 2, id="fail"),
        ),
    )
    def test_run_once(
        default_rules_collection: RulesCollection,
        test_file: str,
        failure: int,
    ) -> None:
        """Test rule matches."""
        results = Runner(test_file, rules=default_rules_collection).run()
        for result in results:
            assert result.rule.id == RunOnce().id
        assert len(results) == failure
